--#!/usr/bin/env luajit
--[[ 
   try to reimplement the exponential formula / Jarzyski
                                  Bennett     / Crooks
   in lua
   fw/bw data files are expected to be first argument after the lua.script, i.e.
                    lua(jit) bar1.lua data.fw data.bw
   TODO: less dumb file reading
--]]

-- catch command line params
local fwfile=arg[1]
local bwfile=arg[2]

-- some error handling
if not fwfile then 
   print("USAGE:")
   print("  lua(jit) " .. arg[0] ..
	 " <forward data> for plain Zwanzig/Jarzynski")
   print("  lua(jit) " .. arg[0] .. " <forward data> <backward data>")
   print("             for Bennett/Crooks")
   os.exit(1)
end

-- "global" variables for rest of this script
local kt=1.987191*300/1000; -- CHARMM units; temp. hardwired
local beta=1/kt;

local function fermi(x)
   -- caller has to make sure that x won't blow math.exp()
   local fx = 1.0/(1.0+math.exp(beta*x))
   return fx
end

local function readdata(input)
   local data = {}
   local file = assert(io.open(input,"r"))
   while true do -- TODO: smarter reading ..
      local val = file:read("*number")
      if not val then break end
      data[#data+1] = val
   end
   file:close()
   return data
end

local function sumdata(data)
   local sum = 0.0
   for i = 1, #data do
      sum = sum + data[i]
   end
   return sum
end

local function meandata(data)
   local mean
   mean = sumdata(data)/#data
   return mean
end

local function vardata(data,mean)
   if (#data < 2) then
      print("Attempting to call variance for less than 2 data points!")
      os.exit(1)
   end
   local var = 0.0
   for i = 1, #data do
      var = var + (data[i] - mean)^2
   end
   return var/#data
end

local function zwanzig(data)
   -- it's the callers responsibility to ensure that math.exp() won't blow up
   local expave = 0.0
   for i = 1, #data do
      expave = expave + math.exp(-data[i]*beta)
   end
   return -kt * math.log(expave/#data)
end

local function minmax(data)
   local maxval = -1.e50
   local minval = -maxval
   for i = 1, #data do
      minval = math.min(minval,data[i])
      maxval = math.max(maxval,data[i])
   end
   return minval, maxval
end

local function bennett(datfw, datbw, toler, maxstep)
   toler = toler or 1.e-10
   maxstep = maxstep or 100
   local csearch = 0.0
   local ratio = #datbw / #datfw
   local ratio_fac = math.log(ratio) * kt
   local iter = 0
   local da_est
   repeat
      iter = iter + 1
      local fwsum = 0
      local bwsum = 0
      for i = 1, #datfw do
	 fwsum = fwsum + fermi(datfw[i]-csearch)
      end
      for i = 1, #datbw do
	 bwsum = bwsum + fermi(datbw[i]+csearch)
      end
      --print(fwsum,bwsum)
      da_est = math.log(bwsum/fwsum) - math.log(ratio) + beta * csearch
      csearch = da_est * kt + ratio_fac
      --print(istep,da_est,csearch,math.abs(fwsum-bwsum))
   until ((math.abs(fwsum-bwsum)<toler) or (iter > maxstep))
   if (iter > maxstep) then
      print("WARNING: did not converge in",maxstep," iterations!")
   end

   da_est = csearch - ratio_fac
   return da_est
end

local function varbennett(datfw,datbw,dabar)
   local ratio = #datbw / #datfw
   local ratio_fac = math.log(ratio) * kt
   local nel = #datfw + #datbw
   local rat1 = nel / #datfw
   local rat2 = nel / #datbw
   local variance = 0
   for i = 1, #datfw do
      variance = variance + 1.0 /
	 (2.0 + 2.0 * math.cosh(beta * ( ratio_fac + datfw[i] - dabar)))
   end
   for i = 1, #datbw do
      variance = variance + 1.0 /
	 (2.0 + 2.0 * math.cosh(beta * (-ratio_fac + datbw[i] - dabar)))
   end
   variance = variance / nel
   variance = 1.0 / ((beta^2)*nel)*(1.0/variance-(rat1+rat2))
   return variance
end

-- Start work ---------------------------------------------------------

local rawfw = {}
local rawbw = {}
local datfw = {}
local datbw = {}

-- reading data
if fwfile then
   rawfw = readdata(fwfile)
else
   print("No forward data file specified -- why was this not caught?")
   os.exit(1)
end
if bwfile then rawbw = readdata(bwfile) end

-- obtain average of forward and -backward data
local offset = sumdata(rawfw)
if bwfile then offset = offset - sumdata(rawbw) end
local nel = #rawfw + #rawbw
local offset = offset/nel

--prepare work data
for i = 1, #rawfw do datfw[i] = rawfw[i] - offset end
if bwfile then
   for i = 1, #rawbw do datbw[i] = rawbw[i] + offset end
end

-- fw exp average
local dafw = zwanzig(datfw)
local meanfw = meandata(datfw)
local varfw = vardata(datfw,meanfw)
local dafw = dafw + offset
local dafwmean = meanfw + offset
local dafwcum  = dafwmean - beta * varfw * 0.5
io.write(string.format("dafw  = %12.3f %12.3f %12.3f\n",dafw,dafwmean,dafwcum))

if bwfile then

   -- bw exp average
   local dabw = zwanzig(datbw)
   local meanbw = meandata(datbw)
   local varbw = vardata(datbw,meanbw)
   local dabw = -(dabw - offset)
   local dabwmean = -(meanbw - offset)
   local dabwcum  = dabwmean + beta * varbw * 0.5
   io.write(string.format("dabw  = %12.3f %12.3f %12.3f\n",dabw,dabwmean,dabwcum))

   -- Bennett

   -- (1) check the pre-requisites
   local fwmin, fwmax = minmax(datfw)
   local bwmin, bwmax = minmax(datbw)
   if (-bwmin < fwmin) then
      print("There seems to be no overlap:")
      print("   fwmin:", fwmin)
      print("  -bwmin:", -bwmin)
      os.exit(1)
   end   

   -- todo: compute (some measure of) overlap info

   -- Bennett free energy estimate
   local dabar
   dabar = bennett(datfw,datbw)
   --dabar = bennett(datfw,datbw,1.e-5,20) -- example of expl. specifying tolerance/max.iterations
   dabar = dabar + offset

   -- OK, using unscaled data and final dabar value; then we get the same
   --     std.dev. as the old fortran code
   -- Bennett intrinsic variance estimate
   local barvar = varbennett(rawfw,rawbw,dabar)
   local stdev = math.sqrt(barvar)

   -- some cumulant approximations
   local dalra = 0.5 * (dafwmean + dabwmean)
   local dalra2 = dalra - (beta/12.0) * (varfw - varbw)

   io.write(string.format("dabar = %12.3f %12.3f %12.3f\n",dabar,dalra,dalra2))
   io.write(string.format("          +- %7.3f\n",stdev))
   
end
